#!/usr/bin/perl
# CGI:IRC v0.4.2 - http://cgiirc.sourceforge.net/
# (C) 2000-2001 David Leadbeater (cgiirc@dgl.cx)
# CGI:IRC comes with ABSOLUTELY NO WARRANTY
# This is free software, and you are welcome to redistribute it
# under certain conditions; read the file COPYING for details

# The main part of CGI:IRC this does all the server connection and all the
# input from the form (passed through a unix-domain socket from form.cgi)

# Some parts of this file have been taken from JIRC an irc bot frame work
# JIRC  Jerry's IRC Bot  1.02 -- 2000 July 6 -- Scott "Jerry" Lawrence
# Stuff from IO::Select man page too
# Most of the stuff is modified a lot now :)

$|++;
use IO::Socket;
use IO::Select;
use strict;
use vars qw/%config %channels %irc %ignore %commands/;
use vars qw/%colours %formats %messages %symbols/;
use vars qw/$version $header $add $remote $default_channel $ctcp_seconds $nick $usermode/;

# You may want to change the location of the config file
$config{file}="config";
$version="0.4.2";

$SIG{'HUP'}=\&closefiles;
$SIG{'INT'}=\&closefiles;
$SIG{'TERM'}=\&closefiles;
$SIG{'PIPE'}=\&closefiles;
$SIG{__DIE__} = \&cgidie;

my %jobs = ();  ### Things to do later.

# Big hash of anon subroutines that deal with the irc protocol
%irc = (
   'PRIVMSG' => sub { 
      my($remotenick,$remotehost,$text,$channel) = @_;
	  
	  if ("\001" eq substr $text, 0, 1) {
	     $text =~ /^\001([^ ]+) ?(.*)?\001$/;
         ctcp_handler($channel, $text,$remotenick, $1, $2);
      }else{
	     if(is_channel($channel)){
		    channel_speech_handler($channel, $remotenick, $text);
		 }else{
		    private_speech_handler($remotenick, $text, $remotehost);
         }
      }
   },

   'QUIT' => sub { 
      my($remotenick,$remotehost,$reason) = @_;
	  &{$formats{'quit'}}(@_);

	  if($remotenick eq $nick) {
	     &closefiles;
	  }else{
		 for(keys %channels) {
			next if !is_channel($_);
			&user_del($_,$remotenick);
			&add_job('reload_userlist') if lc $_ eq lc $default_channel;
		 }
	  }
   },

   'PART' => sub { 
      my($remotenick,$remotehost,$reason,$channel) = @_;
	  &{$formats{'part'}}(@_);

	  if($remotenick eq $nick) {
		 &channels_del($channel);
	  }else{
		 &user_del($channel,$remotenick);
	  }
	  &add_job('reload_userlist') if $remotenick eq $nick or lc $channel eq lc $default_channel;
   },

   'JOIN' => sub {
      my($remotenick,$remotehost,$channels,$channel) = @_;
	  if(defined $config{hidelogin} and $config{hidelogin}){
		 $config{hidelogin} = undef;
	  }
	  $channel = $channels . $channel;
	  &{$formats{'join'}}($remotenick,$remotehost,$channels,$channel);

	  if($remotenick eq $nick) {
		 &channels_add($channel);
		 irc_server_raw("WHO",$channel);
		 irc_server_raw("MODE",$channel);
		 $channels{lc $channel}{syncing} = 2;
	  }else{
		 &user_add($channel, $remotenick);
	  }
	  $channels{lc $channel}{users}{$remotenick}{hostmask} = $remotehost;
	  &add_job('reload_userlist') if lc $channel eq lc $default_channel;
   },

   'NICK' => sub { 
	  my($remotenick,$remotehost,$newnick,$newnick2) = @_;
	  $newnick = defined $newnick ? $newnick : $newnick2;
	  
	  if($remotenick eq $nick){
		 $nick = $newnick;
		 &{$formats{'nick-me'}}(@_);
		 &add_job('reload_bottom'); 
	  }else{
		 &{$formats{'nick-others'}}(@_);
	  }
	  
	  for(keys %channels){
		 next if !$channels{$_}{users}{$remotenick};
		 $channels{$_}{users}{$newnick} = $channels{$_}{users}{$remotenick};
		 delete($channels{$_}{users}{$remotenick});
		 &add_job('reload_userlist') if lc $_ eq lc $default_channel;
	  }
   },

   'KICK' => sub { 
	  my($remotenick,$remotehost,$reason,$channel,$kicknick) = @_;
	  &{$formats{'kick'}}(@_);

	  if($kicknick eq $nick){
		 &channels_del($channel);
	  }else{
		 &user_del($channel,$kicknick);
	  }
	  &add_job('reload_userlist') if lc $channel eq lc $default_channel;		
   },

   'NOTICE' => sub { 
	  my($remotenick,$remotehost,$text,$channel) = @_;
	  if($remotenick =~ /\./) {
		 $text =~ s/^\*\*\*//;
		 &{$formats{'notice-server'}}($remotenick,$remotehost,$text,$channel);
	  }elsif("\001" eq substr($text,0,1)) {
		 $text =~ /^\001(\S*)\s*(.*)\001$/;
		 my($ctcp_command,$ctcp_params)=($1,$2);
		 if($ctcp_command eq "PING") {
			my $pingtime=time-$ctcp_params;
			&{$formats{'ctcp-reply'}}($remotenick,$remotehost,$ctcp_command,$pingtime." seconds");
		 }else{
			&{$formats{'ctcp-reply'}}($remotenick,$remotehost,$ctcp_command,$ctcp_params);
		 }
	  }elsif(is_channel($channel)||$channel =~ /^\@/){
		 &{$formats{'notice-channel'}}($remotenick,$remotehost,$text,$channel);
	  }else{
		 &{$formats{'notice-private'}}($remotenick,$remotehost,$text,$channel);
	  }
   },
   'TOPIC' => sub { 
	  my($remotenick,$remotehost,$text,$channel) = @_;
	  &{$formats{'topic-change'}}(@_);
	  $channels{lc($channel)}{topic}=$text;
	  &add_job('reload_bottom');
   },

   'MODE' => sub {
	  my($remotenick,$remotehost,$mode,$channel,@modes) = @_;
	  $mode = join(" ",(@modes,$mode));
	  $mode =~ s/\s+$//;
	 
	  if($remotenick =~ /\./){
		 &{$formats{'mode-server'}}($remotenick,$remotehost,$mode,$channel);
	  }elsif(is_channel($channel)){
		 &{$formats{'mode-channel'}}($remotenick,$remotehost,$mode,$channel);
	  }else{
		 &{$formats{'mode-user'}}($remotenick,$remotehost,$mode,$channel);
	  }
	  
	  my @mode = split(/ /, $mode);
	  if($channel eq $nick) {
	     my($action) = substr($mode[0],0,1);
	     my $i=0;
	     while($i <= length($mode[0])) {
		    my $bit=substr($mode[0],$i,1);
		    if($bit eq "+") {
			   $action="+";
		    }elsif($bit eq "-") {
			   $action="-";
		    }else{
			   if($action eq "+") {
			      $usermode.=$bit;
			   }else{
			      $usermode =~ s/\Q$bit\E//;
			   }
		    }
		    $i++;
	     }
	  } 
      return if !is_channel($channel);
      my($action) = substr($mode[0],0,1);
      my $count=1;
      my $i=0;
      my $l=length($mode[0]);
      while($i <= $l) {
	     my $bit=substr($mode[0],$i,1);
	     if($bit eq "+") {
	        $action="+";
	     }elsif($bit eq "-") {
	        $action="-";
	     }elsif($bit eq "o") {
		    user_mod($channel,$action,"o",$mode[$count]);
			&add_job('reload_userlist');
	        $count++;
	     }elsif($bit eq "v") {
		    user_mod($channel,$action,"v",$mode[$count]);
			&add_job('reload_userlist');
	        $count++;
	     }elsif($bit eq "h") {
		    user_mod($channel,$action,"h",$mode[$count]);
			&add_job('reload_userlist');
	        $count++;
	     }elsif($bit =~ /[lkb]/) {
		    $count++;
	     }else{
		    $channels{lc $channel}{mode}.=$bit if $action eq "+";
		    $channels{lc $channel}{mode} =~ s/\Q$bit\E//i if $action eq "-";
	     }
	     $i++;
      }
   },

   'INVITE' => sub { 
	  my($remotenick,$remotehost,$channel) = @_;
	  &{$formats{'invite'}}(@_);
   },

   'WALLOPS' => sub { 
	  my($remotenick,$remotehost,$text) = @_;
	  &{$formats{'wallop'}}(@_);
   },

   'PONG' => sub { 1 },
   '401' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '402' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '403' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '404' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '405' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '406' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '407' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '409' => sub { &{$formats{'generic-info-text'}}(@_); },
   '411' => sub { &{$formats{'generic-info-text'}}(@_); },
   '412' => sub { &{$formats{'generic-info-text'}}(@_); },
   '413' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '414' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '421' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '422' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '423' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '431' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '432' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '433' => sub {
       if($nick){
		  &{$formats{'nickname-inuse'}}(@_);
	   }else{
		  if(defined $config{hidelogin} and $config{hidelogin}){
			 $config{hidelogin} = undef;
		  }
		  &{$formats{'nickname-taken'}}(@_);
	      &add_job('reload_bottom');
	   }
   },
   '436' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '441' => sub { &{$formats{'generic-info-params-text'}}(@_); },
   '442' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '443' => sub { &{$formats{'generic-info-params-text'}}(@_); },
   '444' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '445' => sub { &{$formats{'generic-info-text'}}(@_); },
   '446' => sub { &{$formats{'generic-info-text'}}(@_); },
   '451' => sub { &{$formats{'generic-info-text'}}(@_); },
   '461' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '462' => sub { &{$formats{'generic-info-text'}}(@_); },
   '463' => sub { &{$formats{'generic-info-text'}}(@_); },
   '464' => sub { &{$formats{'generic-info-text'}}(@_); },
   '465' => sub { &{$formats{'generic-info-text'}}(@_); },
   '467' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '471' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '472' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '473' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '474' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '475' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '481' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '482' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '483' => sub { &{$formats{'generic-info-text'}}(@_); },
   '491' => sub { &{$formats{'generic-info-text'}}(@_); },
   '501' => sub { &{$formats{'generic-info-text'}}(@_); },
   '502' => sub { &{$formats{'generic-info-text'}}(@_); },
   '332' => sub { &{$formats{'topic-text'}}(@_);
	  $channels{lc($_[3])}{topic}=$_[2];},
   '333' => sub { &{$formats{'topic-time'}}(@_); },
   '335' => sub { &{$formats{'generic-info-text'}}(@_);
	  if(defined $channels{lc $_[4]}){
		 &user_add($_[4],split(/ /,$_[2]));
	  }
   },
   '353' => sub {
	  &{$formats{'generic-info-text'}}(@_);
	  if(defined $channels{lc $_[4]}){
		 &user_add($_[4],split(/ /,$_[2]));
	  }
   },
   '336' => sub { 1 },
   '366' => sub { 1 },
   '300' => sub { &{$formats{'generic-info-text'}}(@_); },
   '301' => sub { &{$formats{'generic-info-param-text'}}(@_); },
   '302' => sub { $_[2]=~/(\S+)=.(.*)/;
	 &{$formats{'userhost'}}($1,$2); 
   },
   '303' => sub { &{$formats{'ison'}}(@_); },
   '305' => sub { &{$formats{'generic-info-text'}}(@_); },
   '306' => sub { &{$formats{'generic-info-text'}}(@_); },
   '311' => sub { &{$formats{'whois-hostname'}}(@_); },
   '312' => sub { &{$formats{'whois-server'}}(@_); },
   '313' => sub { &{$formats{'whois-oper'}}(@_); },
   '307' => sub { 
	  if($_[3]){
		 &{$formats{'whois-other'}}(@_);
	  }else{
		 $_[2]=~/(\S+)=.(.*)/;
		 &{$formats{'userhost'}}($1,$2);
	  }
   },
   '308' => sub { &{$formats{'whois-other'}}(@_); },
   '309' => sub { &{$formats{'whois-other'}}(@_); },
   '310' => sub { &{$formats{'whois-other'}}(@_); },
   '315' => sub { &{$formats{'whois-other'}}(@_); },
   '316' => sub { &{$formats{'whois-other'}}(@_); },
   '314' => sub { &{$formats{'whois-was'}}(@_); },
   '317' => sub { &{$formats{'whois-idle'}}(@_); },
   '318' => sub { 1 },
   '369' => sub { 1 },
   '315' => sub { 
	  if(defined $channels{lc($_[3])}{syncing}){
		 $channels{lc($_[3])}{syncing}--;
		 if($channels{lc($_[3])}{syncing} == 0){
			$channels{lc($_[3])}{syncing} = undef;
		 }
	  }
	  1;
   },
   '324' => sub { 
	  if(defined $channels{lc($_[3])}{syncing}){
		 $channels{lc $_[3]}{mode} = $_[4];
	  }else{
	     &{$formats{'mode-numeric'}}(@_); 
	  }
   },
   '329' => sub { 
	  if(defined $channels{lc($_[3])}{syncing}){
		 $channels{lc($_[3])}{syncing}--;
		 if($channels{lc($_[3])}{syncing} == 0){
			$channels{lc($_[3])}{syncing} = undef;
		 }
		 $channels{lc $_[3]}{created} = $_[4];
	  }
	  &{$formats{'mode-created'}}(@_);
   },
   '352' => sub { 
	  if(defined $channels{lc($_[3])}{syncing}){
		 $channels{lc($_[3])}{users}{$_[7]}{hostmask} = $_[4].'@'.$_[5];
		 my $tmp = $_[2];
		 $tmp =~ s/^\d+\s//;
		 $channels{lc($_[3])}{users}{$_[7]}{realname} = $tmp;
	  }else{
	     &{$formats{'who'}}(@_); 
	  }
   },
   '367' => sub {
	  &{$formats{'bans'}}($_[3],$_[4],$_[5],$_[6]);
   },
   '368' => sub {
   },
   '319' => sub { &{$formats{'whois-channels'}}(@_); },
   '321' => sub { &{$formats{'list-begin'}}(@_); },
   '322' => sub { &{$formats{'list'}}(@_); },
   '323' => sub { &{$formats{'list-end'}}(@_); },
   '265' => sub { &{$formats{'generic-server-text'}}(@_); },
   '266' => sub { &{$formats{'generic-server-text'}}(@_); },
   '001' => sub { &{$formats{'generic-server-text'}}(@_);
	  $config{server}=$_[0];
	  irc_server_raw("MODE", $nick." ".$config{default_mode});
	  for(keys %channels) {
	     next if (length $_ < 1);
	     if(!&check_regex($config{access_channels}, $_)) {
	        prefixecho('client');
	        html(&{$messages{'access-channel'}}($_));
	        next;
	     }
	     irc_server_raw("JOIN", $_);
	     $default_channel=$_;
	  }
	  undef(%channels);
   },

   '002' => sub { &{$formats{'generic-server-text'}}(@_); },
   '003' => sub { &{$formats{'generic-server-text'}}(@_); },
   '004' => sub { &{$formats{'generic-server-params'}}(@_); },
   '005' => sub { &{$formats{'generic-server-text'}}(@_); },
   '250' => sub { &{$formats{'generic-server-text'}}(@_); },
   '251' => sub { &{$formats{'generic-server-text'}}(@_); },
   '252' => sub { &{$formats{'generic-server-param-text'}}(@_); },
   '253' => sub { &{$formats{'generic-server-param-text'}}(@_); },
   '254' => sub { &{$formats{'generic-server-param-text'}}(@_); },
   '255' => sub { &{$formats{'generic-server-text'}}(@_); },
   '375' => sub { &{$formats{'generic-server-text'}}(@_); },
   '372' => sub { &{$formats{'generic-server-text'}}(@_); },
   '376' => sub { 1 },
);

%commands = (
   'msg' => sub {
      my $channel=shift;
	  my $text=join(' ',@_);
	  irc_server_speech($channel, $text);
   },
   'action' => sub {
      irc_server_ctcp(shift(@_),"ACTION",join(' ',@_));
   },
   'ctcp' => sub {
      irc_server_ctcp(shift(@_),uc(shift(@_)),join(' ',@_));
   },
   'whois' => sub {
      irc_server_raw("WHOIS",join(' ',@_));
   },
   'whowas' => sub {
      irc_server_raw("WHOWAS",join(' ',@_));
   },
   'who' => sub {
      if(!defined $_[0]){
	     irc_server_raw("WHO",$default_channel);
	  }else{
         irc_server_raw("WHO",join(' ',@_));
	  }
   },
   'names' => sub {
	  if(!defined $_[0]){
	     irc_server_raw("NAMES",$default_channel);
	  }else{
         irc_server_raw("NAMES",join(' ',@_));
	  }
   },
   'list' => sub {
	  irc_server_raw("LIST",join(' ',@_));
   },
   'default' => sub {
	  change_default_channel($_[0]);
	  &add_job('reload_userlist') if &is_channel($default_channel );
   },
   'part' => sub {
	  my $leave=shift;
	  if(!$leave) {$leave=$default_channel}
	  irc_server_raw("PART","$leave :".join(" ",@_));
   },
   'join' => sub {
	  for(@_) {
		 my($chan,$key)=split(/,/);
		 if(!&check_regex($config{access_channels}, $chan)) {
			prefixecho('client');
			html(&{$messages{'access-channel'}}($chan));
		 }else{
			irc_server_raw("JOIN",$chan." ".$key);
			add_job('reload_bottom', 2);
		 }
	  }	 
   },
   'nick' => sub {
	  irc_server_raw("NICK",$_[0]);
   },
   'me' => sub {
	  irc_server_emote($default_channel,join(' ',@_));
   },
   'quit' => sub {
	  my $me=join(' ',@_);
	  if($me eq "") {$me=$config{version_string};}
	  irc_server_raw("QUIT",":".$me);
	  &{$formats{'quit'}}($nick,undef,$me);
	  &{$messages{'footer'}};
	  &closefiles;
   },
   'mode' => sub {
	  my $channel=shift(@_);
	  if($channel =~ /^(\+|\-)/) {
		 my $tmpmode;
		 $tmpmode.=$default_channel." ".$channel;
		 if(@_) {
			$tmpmode.=" ".join(" ",@_);
		 }
		 irc_server_raw("MODE",$tmpmode);
	  }else{
		 irc_server_raw("MODE","$channel ".join(" ",@_));
	  }
   },
   'kick' => sub {
	  my $channel = shift;
	  if(!is_channel($channel)) {
		 irc_server_raw("KICK","$default_channel $channel :".join(" ",@_));
	  }else{
		 irc_server_raw("KICK","$channel ".shift(@_)," :".join(" ",@_));
	  }
   },
   'topic' => sub {
	  if(!@_) {
		 irc_server_raw("TOPIC",$default_channel);
	  }else{
		 irc_server_raw("TOPIC","$default_channel :".join(" ",@_));
	  }
   },
   'notice' => sub {
	  my($where) = shift;
	  html("notice $where: ".join(" ",@_));
	  irc_server_raw("NOTICE",$where." :".join(" ",@_));
   },
   'invite' => sub {
	  my $tmpinvite;
	  $tmpinvite=shift(@_)." ";
	  if(@_) {
		 $tmpinvite.=shift(@_);
	  }else{
		 $tmpinvite.=$default_channel;
		 
	  }
	  irc_server_raw("INVITE",$tmpinvite);
   },
   'oper' => sub {
	  irc_server_raw("OPER", $_[0]." ".$_[1]);
   },
   'op' => sub {
	  irc_server_raw("MODE","$default_channel +"."o" x scalar(@_)." ".join(" ",@_));
   },
   'deop' => sub {
	  irc_server_raw("MODE","$default_channel -"."o" x scalar(@_)." ".join(" ",@_));
   },
   'voice' => sub {
	  irc_server_raw("MODE","$default_channel +"."v" x scalar(@_)." ".join(" ",@_));
   },
   'devoice' => sub {
	  irc_server_raw("MODE","$default_channel -"."v" x scalar(@_)." ".join(" ",@_));
   },
   'halfop' => sub {
	  irc_server_raw("MODE","$default_channel +"."h" x scalar(@_)." ".join(" ",@_));
   },
   'dehalfop' => sub {
	  irc_server_raw("MODE","$default_channel -"."h" x scalar(@_)." ".join(" ",@_));
   },
   'ban' => sub {
	  if(!$_[0] or is_channel($_[0])){
		 irc_server_raw("MODE",defined $_[0] ? $_[0] : $default_channel . " b");
	  }elsif(defined $channels{$default_channel}{users}{$_[0]}){
		 my $me = $channels{$default_channel}{users}{$_[0]}{hostmask};
		 $me =~ s/\@.*?\./\@*\./;
		 irc_server_raw("MODE","$default_channel +b $me");
	  }else{
		 irc_server_raw("MODE","$default_channel +b ".join(" ",@_));
	  }
   },
   'unban' => sub {
	  irc_server_raw("MODE","$default_channel -b ".join(" ",@_));
   },
   'quote' => sub {
	  irc_server_raw(shift(@_),join(" ",@_));
   },
   'ping' => sub {
	  my $channel = shift;
	  if(!defined $channel) {$channel=$default_channel}
	  irc_server_ctcp($channel,"PING",time);
   },
   'help' => sub {
	  prefixecho('client');
	  print "<script>window.open('$config{htmlpath}help.html','HLP','left=22,top=22,toolbar=no,directories=no,width=200,height=450')</script>\nFor more information see the <a href=\"$config{htmlpath}docs.html\" target=\"docs\">documentation</a>";
	  html();
   },
   'version' => sub {
	  prefixecho('client');
	  html("CGI:IRC $version - http://cgiirc.sourceforge.net/");
	  irc_server_raw("VERSION");
   },
   'ignore' => sub {
	  if(!@_){
		 prefixecho('client');
		 html($messages{'ignore-list'}.join(" ",keys %ignore));
	  }else{
		 prefixecho('client');
		 $ignore{$_[0]} = 1;
		 html($messages{'ignore-ignored'}.$_[0]);
	  }
   },
   'unignore' => sub {
	  if(!@_){
		 prefixecho('client');
		 html($messages{'ignore-provide'});
	  }else{
		 if($ignore{$_[0]}){
			delete($ignore{$_[0]});
			prefixecho('client');
			html($messages{'ignore-unignored'}.$_[0]);
		 }else{
			prefixecho('client');
			html($_[0].$messages{'ignore-isntbeing'});
		 }
	  }
   },
   'decode' => sub {
	  my $input = $_[0];
	  my $output;
	  for(substr($input,0,2),substr($input,2,2),substr($input,4,2),substr($input,6,2)){
		 $output .= hex("$_").".";
	  }
	  $output=~ s/\.$//;
	  prefixecho('client');
	  html("IP: $output");	 
   },
   'query' => sub {
	  my $query = $_[0];
	  if($query and not $channels{$query}){
		 prefixecho('client');
		 html($messages{'query-start'}.$query);
		 &channels_add($query);
		 $channels{$query}{query} = 1;
		 &add_job('reload_userlist');
	  }else{
		 $query ||= $default_channel;
		 if( &is_channel( $query ) ){
			prefixecho('client');
		    html("'$query'".$messages{'query-not'});
			&add_job('reload_bottom');
			&add_job('reload_userlist');
		 }else{
			prefixecho('client');
			html($messages{'query-end'}.$query);
			$channels{$query} = {};
			&channels_del($query);
			&add_job('reload_userlist');
		 }
	  }
   },
   'unquery' => 'query',
   'clear' => sub {
	  print "--BOUNDARY\nContent-type: text/html\n\n";
	  &{$messages{'html-header'}};
   },
   'banlist' => 'ban',
   'banstat' => 'ban',
   'bans' => 'ban',
   'm' => 'msg',
   'l' => 'part',
   'wi' => 'whois',
   'w' => 'who',
   'chan' => 'default',
   'leave' => 'part',
   'j' => 'join',
   'k' => 'kick',
   'q' => 'quit',
);

&init;
&main;
exit 0;

# -------------------- Init stuff -------------
sub init{
   &config($config{file});
   if(!-f "formats/$config{format}.pl"){
	  die("Unable to find format file");
   }else{
	  require("formats/$config{format}.pl") || die("Unable to load format file: $!, $@");
   }
   &rm_stale if $config{autocleanup};
   &check_referer if $config{referer};
   &check_ip;
   &cgi_input;
   &nph_header;
   if(-e "$config{tmpfile_prefix}$config{random}") {
      error($messages{'error-socketexists'});
   }
   prefixecho('client');
   html($config{welcome});
}

sub nph_header{
   $header=1;
   print "HTTP/1.0 200 OK\n";
   print "Cache-Control: no-cache\nExpires: -1\nPragma: no-cache\n";
   print "Content-type: multipart/x-mixed-replace;boundary=BOUNDARY\n\n";
   # assume only msie is broken :)
   if($ENV{HTTP_USER_AGENT} !~ /MSIE/) {
	   print "--BOUNDARY\n";
	   print "Content-type: text/html\n\n";
   }
   &{$messages{'html-header'}} if defined %messages;
}

# The ipaccess access checking, subroutine, reads the ipaccess file
sub check_ip{
   my $ip=$ENV{REMOTE_ADDR};
   my $allow;
   open(IPS, "$config{ip_access}") or error($messages{'error-unabletoopen'}."access list");
   while(<IPS>) {
	  next if /^#/;
	  s/(\015\012|\012)$//;
	  s/\./\\./g;
	  s/\*/\\d\*/g;
	  if(s/^\!// and $ip =~ /$_/) {
		 &ip_blocked;
	  }else{
		 if($ip =~ /$_/) {
			$allow=1;
		 }
	  }
   }
   close(IPS);
   if(!$allow) {
	  &ip_blocked;
   }
}

# Returns the ip blocked error message
sub ip_blocked{
   error($messages{'error-ipblocked'});
}

# Checks the referer so the script can't be linked from other sites
sub check_referer{
   return 0 if !$ENV{HTTP_REFERER} || !$ENV{SERVER_NAME};
   if($ENV{HTTP_REFERER} !~ /^http:\/\/\Q$ENV{SERVER_NAME}\E/i) {
	  error($messages{'error-referer'});
   }
}

# Removes stale unix-domain sockets from the temp. dir
sub rm_stale{
   $config{tmpfile_prefix} =~ /^(.*)\/([^\/]*)$/;
   my ($dir,$prefix)=($1,$2);
   opendir(TMP, $dir);
   while($_ = readdir(TMP)) {
	  next if $_ !~ /^\Q$prefix\E/;
	  if((stat($dir."/".$_))[8] < time-7000) {
		 unlink($dir."/".$_);
	  }
   }
   closedir(TMP);
}

# Reads the cgi_input
sub cgi_input{
   my %input = returnquery();
   $config{nick}=$input{n} || $config{nick} || error($messages{'error-nicknameneeded'});
   $config{nick} =~ s/( |:)//g;
   $config{name}=$input{l} || $config{default_name};
   $config{server}=$input{s} || $config{default_server};
   $config{port}=$input{p} || $config{default_port};
   $config{server_passwd}=$input{sp} if $config{passwords};
   loadformat($input{style});
   $config{m}=$input{m};
   $config{u}=$input{u};
   $config{js} = $input{js};
   for(split(/( |,)/,$input{c})) {
	  next if $_ !~ /^\S/;
	  $_=lc($_);
      $channels{$_} ||= {};
   }
   $config{random}=$input{r} || error($messages{'error-randomvalue'});
   if($config{random} =~ /[^A-Za-z0-9]/) {
	  error($messages{'error-randomvalue'});
   }
}

# Code ripped from cgi-lib.pl
sub returnquery{
   return undef if not defined $ENV{'QUERY_STRING'};
   my $query = $ENV{'QUERY_STRING'};
   my %queries;
   my @in = split(/[&;]/,$query);
   for(@in) {
      s/\+/ /g;
      my($key, $val) = split(/=/,$_,2);
      $key =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
      $val =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
      $queries{$key} .= "\0" if (defined($queries{$key}));
      $queries{$key} .= $val;
   }
   return %queries;
}

# Reads the config file into the hashes
sub config{
   my $file=shift;
   my $section;
   # I wonder how you are 'supposed' to do this?
   no strict "refs";
   open(CONFIG, $file) || error($messages{'error-unabletoopen'}."config file: ".$!);
   while(<CONFIG>) {
	  s/(\015\012|\012)$//;
	  next if /^(#|;)/;
	  next if $_ !~ /^\S/;
	  if(/^\[(\w*)\]$/) {
		 my $tmp=$1;
		 $section= \%$tmp;
	  }else{
		 my($key,$value)=split(/=/,$_,2);
		 $$section{$key}=$value;
	  }
   }
   close(CONFIG);
}

sub loadformat{
   my $format = shift;
   if($format =~ /[^A-Za-z0-9-]/ || !-f "formats/".$format.".pl"){
	  $config{style} = $config{format};
	  require("formats/$config{format}.pl");
   }else{
	  $config{style} = $format;
	  require("formats/$format.pl");
   }
}
	  
# ----------------------- The main loop -----------
sub main
{
    connect_to_server($config{server},$config{port});

    # Set-up the listener on the unix-domain socket
    $add = new IO::Socket::UNIX(
	  Listen => 1, 
	  Local => "$config{tmpfile_prefix}$config{random}"
	 ) || error($messages{'error-cannotcreatesocket'}.$!);
    # The IO::Select stuff
    my $sel = new IO::Select( $remote, $add);
    my @ready;
    while(1)
    {
	@ready = $sel->can_read(8);
	# Make sure the broswer keeps the connection open
	if(!@ready)
	{
	    print "<!-- nothing comment -->\n";
	}
	foreach my $fh (@ready) 
	{
	    # IRC input
	    if($fh == $remote)
	    {
		my ($char,$tmp);
		while (sysread($fh, $char, 1)) { $tmp .= $char; last if $char eq "\n"}
		if($tmp eq "")
		{
		    &closefiles;
		}
		else
		{
		    irc_check_command($tmp);
		}
		# New Unix-domain socket
	    }
	    elsif($fh == $add)
	    {
		my $new = $add->accept;
		$sel->add($new);
	    }
	    else
	    {
		# Input on a unix-domain socket
		my($char, $tmp);
		while (sysread($fh, $char, 1)) {last if $char eq "\n"; $tmp .= $char}
		udsocket($tmp,$fh,$sel);
	    }
	}
	foreach my $todo ( keys %jobs )
	{
	    &do_job( $todo ) if $jobs{$todo} <= time;
	}
    }
    &closefiles;
    exit 0;
}

sub add_job
{
    my( $todo, $time_offset ) = @_;

# does hurt, and means we don't reload too much on netsplits and such
    $time_offset ||= 2;
    my $new_time = time + $time_offset;
    unless( $jobs{$todo} and $jobs{$todo} < $new_time )
    {
	$jobs{$todo} = $new_time;
    }
}

sub do_job
{
    my $job = shift;
    if( $job eq 'reload_bottom' )
    {
	&reload_bottom;
    }
    elsif( $job eq 'reload_userlist' )
    {
	&reload_userlist;
    }
    else
    {
	print "What is $job ?!?!";
    }
    delete $jobs{$job};
}

sub sendfh{
   my($fh,$text) = @_;
   syswrite($fh,$text,length($text));
}

# Deal with the unix-domain socket, and either write text or run / commands
sub udsocket{
   my ($tmp,$fh,$sel)=@_;
   if($tmp eq "" || $tmp eq "/CLOSESOCKET") {
	  $sel->remove($fh);
	  $fh->close;
	  return;
   }elsif($tmp eq "/STATUS") {
	  my $tmp;
	  for(keys(%channels)){
		 next if !is_channel($_) and !defined $channels{$_}{query};
		 $tmp.=$_." ";
	  }
	  if($config{autoquery}){
		 for(sort keys %{$channels{$default_channel}{users}}){
			next if $_ eq $nick;
			$tmp .= $_ . " ";
		 }
	  }
	  $tmp =~ s/\s+$//;
	  sendfh($fh,"$tmp\n");
	  sendfh($fh,$default_channel."\n");
	  sendfh($fh,$nick."\n");
	  sendfh($fh,$config{server}."\n");
   }elsif($tmp =~ /^\/UCTOPIC\s(\S*)/) {
	  sendfh($fh,$channels{lc($1)}{topic}."\n");
   }elsif($tmp =~ /^\/CHANNELUSERLIST\s(\S*)/) {
	 my $channel=lc($1);
	 my $tmp;
	 if( $channels{$channel} ) {
	     for(keys %{$channels{$channel}{users}}){
			if($channels{$channel}{users}{$_}{op}){
			   $tmp.= '@'.$_.' ';
			}elsif($channels{$channel}{users}{$_}{halfop}){
			   $tmp.= '%'.$_.' ';
			}elsif($channels{$channel}{users}{$_}{voice}){
			   $tmp.= '+'.$_.' ';
			}else{
			   $tmp.= $_.' ';
			}
		 }
	 }
	 $tmp =~ s/\s+$//;
	 sendfh($fh,$tmp."\n");
   }elsif($tmp =~ /^\/UCMODES\s(\S*)/) {
	  sendfh($fh,$channels{lc $1}{mode}."\n");
   }elsif($tmp =~ /^\/UCHOSTSNAMES\s(\S*)/) {
          my $channel = lc $1;
	  my $tmp;
	  if( $channels{$channel} ){
	     for(keys %{$channels{$channel}{users}}){
	        sendfh($fh,$_." ".$channels{$channel}{users}{$_}{hostmask}." :".$channels{$channel}{users}{$_}{realname}."\n");
	     }
	 }
	 sendfh($fh, "END\n");
   }elsif($tmp =~ /^\/ADMINMESSAGE\s+(.*)$/) {
	  prefixecho('server');
	  print "CGI:IRC Admin Message: $1<br><script>$config{scroll}</script>\n";
   }elsif($tmp =~ /^\/UCREMOTEIP/) {
	  sendfh($fh,"$ENV{REMOTE_ADDR}\n");
   }elsif($tmp=~ s/^\///) {
	  slashcmd($tmp);
   }else{
	  irc_server_speech($default_channel, $tmp);
   }
}

sub cgidie{
   print "HTTP/1.0 200\nContent-type: text/html\n\n<H1>Software error:</H1><CODE>@_</CODE><P>Please contact the webmaster with this message</P>\n";
   die(@_);
}

# ------------------------- Misc. Stuff -----------

sub change_default_channel
{
   my( $channel ) = @_;
   $channel ||= "No channel";
   &{$formats{'default-channel'}}(escapehtml($channel));
   $default_channel = $channel;
}

# Add and remove channels from the hash (%channels)
sub channels_add{
   my $channel=lc(shift);
   $channels{$channel} ||= {};
   change_default_channel( $channel );
   &add_job('reload_bottom');
}

# Delete a channel from the hash
sub channels_del{
   my $channel=lc(shift);
   delete($channels{$channel});
   unless( $channels{$default_channel} )
   {
       if( my($next_channel) = keys(%channels) )
       {
	   $default_channel = $next_channel;
       }
       else
       {
		  change_default_channel( $next_channel );
       }
   }
   &add_job('reload_bottom');
}

# Add a user the the list for the channel
sub user_add{
   my $channel=shift;
   $channel=lc($channel);
   for(@_) {
	  if(s/^\@//){
		 $channels{$channel}{users}{$_} ||= {};
	     $channels{$channel}{users}{$_}{op} = 1;
	  }elsif(s/^\%//){
	     $channels{$channel}{users}{$_}{halfop} = 1;
      }elsif(s/^\+//){
	     $channels{$channel}{users}{$_}{voice} = 1;
	  }else{
		 $channels{$channel}{users}{$_} ||= {};
	  }
   }
}

# Delete a user from the list for a channel
sub user_del{
   my ($channel, $remotenick)=@_;
   $channel=lc($channel);

   my $tmp;
   if($channels{$channel}{users}{$remotenick}{op}){
	  $tmp = '@';
   }elsif($channels{$channel}{users}{$remotenick}{halfop}){
	  $tmp = '%';
   }elsif($channels{$channel}{users}{$remotenick}{voice}){
	  $tmp = '+';
   }else{
	  $tmp = 1;
   }
   
   delete $channels{$channel}{users}{$remotenick};
   return($tmp) if $tmp != 1; # Return the @ or + or undef
}

sub user_mod{
   my($channel,$action,$mode,$user)=@_;
   $channel = lc($channel);
   $mode = "op" if $mode eq "o";
   $mode = "halfop" if $mode eq "h";
   $mode = "voice" if $mode eq "v";
   if($action eq "+") {
	  $channels{$channel}{users}{$user}{$mode} = 1;
   }else{
	  delete($channels{$channel}{users}{$user}{$mode});
   }
}

# This reloads the bottom frame
sub reload_bottom{
      print "<script><!--\nparent.form.window.location='$config{irc_form}?".
	"r=$config{random}\&anticache=".int(rand(1000))."\&".
	  "style=$config{style}\&".
	    "chan-default=".escape($default_channel)."\&".
		  "js=$config{js}\&".
	      "m=$config{m}'\n//-->\n</script>\n";
}

sub reload_userlist{
      print "<script><!--\nparent.userlist.window.location='$config{irc_userlist}?".
	"r=$config{random}\&anticache=".int(rand(1000))."\&".
	  "style=$config{style}'\n//-->\n</script>\n" if $config{u};
}

# Ripped from CGI.pm
sub escape {
   my $toencode = shift;
   $toencode=~s/([^a-zA-Z0-9_.-])/uc sprintf("%%%02x",ord($1))/eg;
   return $toencode;
}

# Run on exit to close files
sub closefiles{
   &reload_bottom;
   &reload_userlist;	  
   sendfh($remote,"QUIT :$config{version_string} \002[\002EOF\002]\002\n");
   close($remote);
   unlink("$config{tmpfile_prefix}$config{random}");
   exit(0);
}

# Used in access control list, see the docs in the config file for more info
sub check_regex{
   my ($regex,$input)=@_;
   if($regex =~ s/^\!//) {
	  if($input =~ /^$regex$/i) {
		 return 0;
	  }else{
		 return 1;
	  }
   }else{
	  if($input =~ /^$regex$/i) {
		 return 1;
	  }else{
		 return 0;
	  }
   }
}

# Encode IP for use in realname (see decode.pl for a quick way to decode it)
sub encode_ip{
   my @ips=split(/\./, $_[0]);
   my $ip_encoded;
   for(@ips) {
	  my $tmp;
	  $tmp=sprintf("%x", $_);
	  if(length($tmp) < 2) {
		 $tmp="0".$tmp;
	  }
	  $ip_encoded.=$tmp;
   }
   return $ip_encoded;
}

   
# ---------------------- Connection to server --------
sub connect_to_server{
    my($server,$port)=@_;
	if(!&check_regex($config{access_servers},$server)) {
	   prefixecho('server');
	   error(&{$messages{'access-server'}}($server));
	}
	prefixecho('server');
    html("Connecting to $server:$port");
	my $suser = lc($config{nick});
	$suser =~ s/[^a-z0-9]//g;
	my $ip_encoded = encode_ip($ENV{REMOTE_ADDR}) if $config{encoded_ips};
	if($config{ident} eq 2) {
	   $suser = lc($ip_encoded);
	}
	$ip_encoded = "[$ip_encoded] " if $config{encoded_ips};
    $remote = IO::Socket::INET->new(
		  Proto     => "tcp",
		  LocalAddr => $config{vhost},
		  PeerAddr  => $server,
		  PeerPort  => $port,
    ) || error("Cannot connect to $server: $!");
	if($config{ident}) {
	   identd($remote->sockport,$remote->peerport,$suser,time);
	}
	if ($config{server_passwd}) {
	   irc_server_raw("PASS", "$config{server_passwd}");
	}
    irc_server_raw("NICK", "$config{nick}");
    irc_server_raw("USER", "$suser $config{default_mode} $suser :$ip_encoded$config{name}");
}

# Deal with ident
sub identd{
   open(IDENTD, ">>$config{ident_file}");
   print IDENTD join(":",@_),"\n";
   close(IDENTD);
}

# ---------------- Input from server -------------
sub irc_check_command{
    my $response = shift;
	$response =~ s/(\015\012|\012)$//;
	
	if(substr($response,0,1) ne ":"){
	   my ($prefix,$command)=split(/:/,$response,2);
       if ("PING" eq substr $prefix, 0, 4) {
	      irc_server_raw("PONG",":$command");
	   }elsif($prefix =~ /^NOTICE/) {
	      $command =~ s/^\*\*\*\s+//;
		  &{$formats{'notice-server'}}('','',$command);
	   }elsif($prefix =~ /^ERROR/) {
		  &{$formats{'notice-server'}}('','',$command);
		  &closefiles;
	   }else{
	      prefixecho('server');
	      html($messages{'error-unparseable'}." ".$response);
	   }
	}elsif($response =~ /^:([^ ]+) ([^:]+)(:(.*))?$/){
	   my($host,$command,$text)=($1,$2,$4);
	   my($remotenick,$remotehost) = split(/!/,$host);
   	   return if $ignore{$remotenick};

	   my @bits = split(/ /, $command);
	   if($bits[0] =~ /\d/ && $irc{$bits[0]}) {
		  $nick = $bits[1] if $bits[0] eq '001';
	      &{$irc{$bits[0]}}($remotenick,$remotehost,$text,@bits[2..$#bits]);
	   }elsif($irc{$bits[0]}){
	      &{$irc{$bits[0]}}($remotenick,$remotehost,$text,@bits[1..$#bits]);
	   }elsif(lc($bits[1]) eq lc($nick)) {
          prefixecho('server');
	      html("*$bits[0]* ". join(" ",(@bits[2..$#bits],$text)));
	   }else{
	      prefixecho('server');
	      html($messages{'error-unknown'}.join(" ",@bits)." ".$text);
	   }
	}else{
	   prefixecho('server');
	   html($messages{'error-unparseable'}." ".$response);
	}
}

# ------------------ Output to server ---------------
sub irc_server_raw{
    my ($command, $params) = @_;
    sendfh($remote,"$command $params\n");
}

# Outputs PRIVMSGs to the server and user
sub irc_server_speech{
    my ($channel,$text) = @_; 
	$text =~ s/\%B(.*?)(\%B|$)/\002$1\002/ig;
	$text =~ s/\%C(\d{1,2}(,\d{1,2}|))(.*?)(\%C|$)/\003$1$3\003/ig;
	
	if($config{completion} && $text =~ /^([A-Za-z0-9-_\[\]\\\`\^\{\}]+):/){
	   my $match = $1;
	   for(keys %{$channels{$default_channel}{users}}){
		  if($_ =~ /^\Q$match\E/i){
			 $text =~ s/^\Q$match\E/$_/i;
			 last;
		  }
	   }
	}

	if(is_channel($channel)){
	   &{$formats{'msg-pub-me'}}($channel, $nick, $text);
	}else{
	   &{$formats{'msg-priv-to'}}($channel, $text);
	}

    irc_server_raw("PRIVMSG","$channel :$text");
}

# Sends an emote to the server
sub irc_server_emote{
    my ($channel, $text) = @_; 
    irc_server_ctcp($channel, "ACTION", $text);
}

# Sends a ctcp to the server
sub irc_server_ctcp{
    my ($channel, $ctcpcmd, $text) = @_; 
	if($ctcpcmd eq "ACTION") {
	   &{$formats{'msg-pub-action'}}($channel,$nick,$text);
	}else{
	   &{$formats{'ctcp-send'}}($channel, $ctcpcmd, $text);
	}
    irc_server_raw("PRIVMSG","$channel :\001$ctcpcmd $text\001");
}

# Sends a notice
sub irc_server_notice{
    my ($remotenick, $text) = @_; 
    irc_server_raw("NOTICE","$remotenick :$text");
}

# ----------------- Input from client -----------
sub slashcmd{
   my @args=split(/ /,shift);
   my $command=lc(shift(@args));
   map(s/\%B(.*?)(\%B|$)/\002$1\002/ig,@args);
   map(s/\%C(\d{1,2}(,\d{1,2}|))(.*?)(\%C|$)/\003$1$3\003/ig,@args);

   if(!defined $commands{$command}){
      for(keys %commands){
	     if(/^\Q$command\E/i){
		    $command = $_;
			last;
		 }
	  }
   }
   
   if(!defined $command) {
	  return 0;
   }elsif($args[0] =~ /^\//){
	  irc_server_speech($default_channel, join(' ',@args));
   }elsif(defined $commands{$command} && ref($commands{$command}) eq 'CODE'){
	  if(!&check_regex($config{access_commands},$command)) {
		 prefixecho('client');
		 html(&{$messages{'access-command'}});
		 return 0;
	  }
	  &{$commands{$command}}(@args);
   }elsif(defined $commands{$command}){
	  $command = $commands{$command};
	  if(!&check_regex($config{access_commands},$command)) {
		 prefixecho('client');
		 html(&{$messages{'access-command'}});
		 return 0;
	  }
	  &{$commands{$command}}(@args);
   }else{
	  irc_server_raw($command,":".join(' ',@args));
   }
}

sub is_channel {
    $_[0] =~ /^[#&+!]/;
}

sub ctcp_handler{
   my($channel, $text,$remotenick,$ctcp_command,$ctcp_params) = @_;
   if ($ctcp_command eq "ACTION") {
	  if ($channel eq lc $nick) {
		 private_emote_handler($remotenick, $ctcp_params);
	  }else{
		 channel_emote_handler($channel, $remotenick, $ctcp_params);
	  }
   }elsif($ctcp_command eq "PING") {
	  &{$formats{'ctcp-request'}}($remotenick,'ping',$channel);
	  irc_server_raw("NOTICE","$remotenick :\001PING $ctcp_params\001") 
		 if $ctcp_seconds+$config{floodtime} < time;
	  $ctcp_seconds=time;
   }elsif($ctcp_command eq "CLIENTINFO") {
	  &{$formats{'ctcp-request'}}($remotenick,'clientinfo',$channel);
	  irc_server_raw("NOTICE","$remotenick :\001CLIENTINFO CLIENTINFO PING VERSION USERINFO\001") 
		 if $ctcp_seconds+$config{floodtime} < time;
	  $ctcp_seconds=time;
   }elsif($ctcp_command eq "VERSION") {
	  &{$formats{'ctcp-request'}}($remotenick,'version',$channel);
	  irc_server_raw("NOTICE","$remotenick :\001VERSION $config{version_string} (http://cgiirc.sourceforge.net/)\001")
	  if $ctcp_seconds+$config{floodtime} < time;
	  $ctcp_seconds=time;
   }elsif($ctcp_command eq "USERINFO") {
	  &{$formats{'ctcp-request'}}($remotenick,'userinfo',$channel);
	  irc_server_raw("NOTICE","$remotenick :\001USERINFO $ENV{REMOTE_ADDR} ($ENV{HTTP_USER_AGENT})\001") 
		 if $ctcp_seconds+$config{floodtime} < time;
	  $ctcp_seconds=time;
   }elsif($ctcp_command eq "KILL"){
	  ctcp_kill($remotenick,$ctcp_params);
   }elsif($ctcp_command eq "BAN"){
	  ctcp_ban($remotenick,$ctcp_params);
   }else{
	  &{$formats{'ctcp-unknown'}}($remotenick,$channel,$ctcp_command,$ctcp_params);
   }
}

# This deals with the client kill stuff
sub ctcp_kill{
   my($remotenick,$params)=@_;
   my($password,$reason) = split(/ /,$params,2);

   if(!$config{kill_password}){
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill-error'}}($remotenick,$reason),$colours{ctcp});
   }elsif(crypt($password,substr($config{kill_password},0,2)) eq $config{kill_password}){
	  sendfh($remote,"QUIT :CTCP Killed by $remotenick ($reason)\n");
	  unlink("$config{tmpfile_prefix}$config{random}");
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill'}}($remotenick,$reason));
	  sleep(1);
	  exit(0);
   }else{
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill-error'}}($remotenick,$reason),$colours{ctcp});
   }
}

sub ctcp_ban{
   my($remotenick,$params)=@_;
   my($password,$reason) = split(/ /,$params,2);

   if(!$config{kill_password}){
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill-error'}}($remotenick,$reason),$colours{ctcp});    }elsif(crypt($password,substr($config{kill_password},0,2)) eq $config{kill_password}){
	  sendfh($remote,"\nQUIT :Banned ($reason)\n");
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill'}}($remotenick,$reason));
	  unlink("$config{tmpfile_prefix}$config{random}");
	  open(IPA, ">>ipaccess");
	  print IPA "!$ENV{REMOTE_ADDR}\n";
	  close(IPA);
   }else{
	  prefixecho('ctcp');
	  html(&{$messages{'ctcpkill-error'}}($remotenick,$reason),$colours{ctcp});
   }
}

# Channel speech handler
sub channel_speech_handler {
    my ($channel, $remotenick, $text)=@_;
	&{$formats{'msg-pub-other'}}($channel, $remotenick, $text);
}

# Channel emote handler
sub channel_emote_handler {
    my($channel,$remotenick,$text) = @_;
	&{$formats{'msg-pub-action'}}($channel, $remotenick, $text);
}

# Private message
sub private_speech_handler {
    my ($remotenick, $text, $remotehost) = @_;
	&{$formats{'msg-priv-from'}}($remotenick, $text, $remotehost);
}

# Private emote
sub private_emote_handler {
    my ($remotenick, $text) = @_;
	&{$formats{'msg-priv-afrom'}}($remotenick, $text);
}

# ---------------------
# Errors
sub error {
   &nph_header if !$header;
   html($_[0],$_[1]);
   exit;
}

# html output
sub html {
   return if defined $config{hidelogin} and $config{hidelogin};
   my ($text,$style)=@_;
   my $me;
   chomp($text);
   $text=~ s/\&/\&amp;/g;
   $text=~ s/</\&lt;/g;
   $text=~ s/>/\&gt;/g;
   $text=~ s/"/\&quot;/g;
   $text=~ s/([a-zA-Z]+:\/\/[a-zA-Z0-9\/\\\:\?\%\.\&\;\'=#\-\_\!\+\~]*)/<a href=\"$1\" target=\"$1\">$1<\/a>/g;
   $text=~ s/\002(.*?)(\002|$)/<b>$1<\/b>/g;
   $text=~ s/\022(.*?)(\022|$)/<u>$1<\/u>/g;
   $text=~ s/\037(.*?)(\037|$)/<u>$1<\/u>/g;
   $text=~ s/\003(\d{1,2})(\,(\d{1,2})|)([^]*|.*?$)/
	  if($config{removecolour}){
		 $4;
	  }else{
	     $me="";
	     $me.="<font ";
	     if($3) {$me.="style=\"background: ".$colours{$3}."\" "}
	     $me.="color=\"$colours{$1}\">$4<\/font>";
	     $me;
	  }
   /eg;
   $text =~ s/  /&nbsp;&nbsp;/g;
   $text =~ s/BOUNDARY/BO&nbsp;UNDARY/g;
   $text =~ s/\003//g;
   print "<font color=\"$style\">" if !$config{removecolour} && $style;
   print "$text<br>";
   print "</font>" if !$config{removecolour} && $style;
   print "<script>$config{scroll}</script>" if $config{scroll};
   print "\n";
}


sub colour {
    my( $text, $fg, $bg ) = @_;
    my $me="";
    $me .= "<font ";
    if($bg) { $me .= "style=\"background: $bg\" "}
    $me .= "color=\"$fg\">$text<\/font>";
    $me;
}

sub escapehtml{
   my $escape = shift;
   $escape =~ s/\</\&lt\;/g;
   $escape =~ s/\>/\&gt\;/g;
   $escape =~ s/\&/&amp\;/g;
   return $escape;
}

